require"inotify"

inot = assert(inotify.init(true))
assert(inot:fd() >= 0)
i,j = string.find(tostring(inot), 'inotify 0x%x+') assert(i==1 and j==tostring(inot):len())
assert(inot:close() == nil)
assert(pcall(function() inot:close() end) == false)

inot = assert(inotify.init(true))
wd, errno, errstr = inot:add_watch('/like/a/skein/of/loose/silk')
assert(wd == nil and errno and errstr)			-- ENOENT
assert(pcall(function() inot:add_watch('.', 'IN_ALL_EVENTS') end) == false)
assert(pcall(function() inot:add_watch('.', {'xxx'}) end) == false)
errno, errstr = inot:rm_watch(99)
assert(errno and errstr)
assert(inot:close() == nil)

local osexecute = function(cmd) assert(os.execute(cmd) == 0) end

local checkevent = function(event, eevent)
	assert(event.wd == eevent.wd)
	event.name = event.name:gsub('%z+', '')		-- remove trailing NULs
	assert(event.name == eevent.name)
	local maskevents = inotify.getmaskevents(event.mask)
	assert(#maskevents == #eevent.events)
	local events={} for i,v in ipairs(maskevents) do events[v] = true end
	for i,v in ipairs(eevent.events) do assert(events[v] == true) end
end

local checkcmdevents = function(cmd, eevents, nonzerocookies)
	osexecute(cmd)
	local nread, errno, errstr = inot:nread()
	assert(nread > 0 and errno == nil and errstr == nil)
	local events, nread, errno, errstr = inot:read()
	assert(#events == #eevents and nread > 0 and errno == nil and errstr == nil)
	for i,v in ipairs(events) do checkevent(v, eevents[i]) end
	local cookie = events[1].cookie
	assert((cookie ~= 0) == (nonzerocookies == true))
	for i,v in ipairs(events) do assert(v.cookie == cookie) end
end

inot = assert(inotify.init(true))

osexecute('rm -rf testdir; mkdir testdir')
wd, errno, errstr = inot:add_watch('testdir')
assert(wd >= 0 and errno == nil and errstr == nil)

wd2, errno, errstr = inot:add_watch('testdir', {'IN_OPEN', 'IN_ONLYDIR', 'IN_MASK_ADD'})
assert(wd2 == wd and errno == nil and errstr == nil)

events, nread, errno, errstr = inot:nbread()
assert(events == nil and nread == nil and errno == nil and errstr == nil)

checkcmdevents('touch testdir', {{wd=wd, name='', events={'IN_ATTRIB', 'IN_ISDIR'}}})

checkcmdevents('cp Makefile testdir/foo', {
	{wd=wd, name='foo', events={'IN_CREATE'}},
	{wd=wd, name='foo', events={'IN_OPEN'}},
	{wd=wd, name='foo', events={'IN_MODIFY'}},
	{wd=wd, name='foo', events={'IN_CLOSE_WRITE'}}
})

checkcmdevents('touch testdir/foo', {
	{wd=wd, name='foo', events={'IN_OPEN'}},
	{wd=wd, name='foo', events={'IN_ATTRIB'}},
	{wd=wd, name='foo', events={'IN_CLOSE_WRITE'}}
})

checkcmdevents('cat testdir/foo >/dev/null', {
	{wd=wd, name='foo', events={'IN_OPEN'}},
	{wd=wd, name='foo', events={'IN_ACCESS'}},
	{wd=wd, name='foo', events={'IN_CLOSE_NOWRITE'}}
})

checkcmdevents('cp /dev/null testdir/foo', {
	{wd=wd, name='foo', events={'IN_MODIFY'}},	-- why is this before IN_OPEN???
	{wd=wd, name='foo', events={'IN_OPEN'}},
	{wd=wd, name='foo', events={'IN_CLOSE_WRITE'}}
})

checkcmdevents('mv testdir/foo .', {{wd=wd, name='foo', events={'IN_MOVED_FROM'}}}, true)
checkcmdevents('mv foo testdir', {{wd=wd, name='foo', events={'IN_MOVED_TO'}}}, true)

checkcmdevents('mv testdir/foo testdir/bar', {
	{wd=wd, name='foo', events={'IN_MOVED_FROM'}},
	{wd=wd, name='bar', events={'IN_MOVED_TO'}},
}, true)
checkcmdevents('mv testdir/bar testdir/foo', {
	{wd=wd, name='bar', events={'IN_MOVED_FROM'}},
	{wd=wd, name='foo', events={'IN_MOVED_TO'}},
}, true)

checkcmdevents('rm -f testdir/foo', {{wd=wd, name='foo', events={'IN_DELETE'}}})
checkcmdevents('mv testdir testdir.mv', {{wd=wd, name='', events={'IN_MOVE_SELF'}}})
checkcmdevents('mv testdir.mv testdir', {{wd=wd, name='', events={'IN_MOVE_SELF'}}})

checkcmdevents('rmdir testdir', {
	{wd=wd, name='', events={'IN_DELETE_SELF'}},
	{wd=wd, name='', events={'IN_IGNORED'}}
})

errno, errstr = inot:rm_watch(wd)
assert(errno and errstr)				-- EINVAL

osexecute('rm -rf testdir; mkdir testdir')
osexecute('touch testdir/foo')
osexecute('cd testdir && ln -s foo bar')

wdfoo, errno, errstr = inot:add_watch('testdir/foo')
assert(wdfoo >= 0 and errno == nil and errstr == nil)
wdbar, errno, errstr = inot:add_watch('testdir/bar')
assert(wdbar >= 0 and errno == nil and errstr == nil)
assert(wdfoo == wdbar)

wdbar, errno, errstr = inot:add_watch('testdir/bar', {'IN_ALL_EVENTS', 'IN_DONT_FOLLOW'})
assert(wdbar >= 0 and errno == nil and errstr == nil)
assert(wdfoo ~= wdbar)

errno, errstr = inot:rm_watch(wdfoo)
assert(errno == nil and errstr == nil)
errno, errstr = inot:rm_watch(wdbar)
assert(errno == nil and errstr == nil)

assert(inot:close() == nil)
osexecute('rm -rf testdir')
